package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.network.chat.Component
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.ForgeHooks
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.IEnergyStorage
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.menu.ChemicalGeneratorMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.lang.ref.WeakReference

class ChemicalGeneratorBlockEntity(pos: BlockPos, state: BlockState) : MatteryBlockEntity(MBlockEntities.CHEMICAL_GENERATOR, pos, state) {
	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return ChemicalGeneratorMenu(containerID, inventory, this)
	}

	private var valid = true

	@JvmField
	val energy = GeneratorEnergyStorage(this::setChangedLight, MAX_ENERGY, THROUGHPUT)

	override fun setChangedLight() {
		super.setChangedLight()
		check = true
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid && (cap == MatteryCapability.ENERGY || cap == ForgeCapabilities.ENERGY) && side !== blockState.getValue(RotatableMatteryBlock.FACING))
			return energy.resolver.cast()

		if (valid && cap == ForgeCapabilities.ITEM_HANDLER)
			return itemHandler.get().cast()

		return super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		itemHandler.invalidate()
		energy.invalidate()
		valid = false
	}

	override fun reviveCaps() {
		super.reviveCaps()
		itemHandler.revive()
		energy.revive()
		valid = true
	}

	override fun setLevel(level: Level) {
		super.setLevel(level)
		tickOnceServer(this::checkSurroundings)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		nbt["energy"] = energy.serializeNBT()
		nbt["container"] = container.serializeNBT()
		nbt["working_ticks"] = workingTicks
		nbt["working_ticks_total"] = workingTicksTotal
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("energy", CompoundTag::class.java, energy::deserializeNBT)
		container.deserializeNBT(nbt["container"])

		nbt.ifHas("working_ticks") {
			if (it is IntTag)
				workingTicks = it.asInt
		}

		nbt.ifHas("working_ticks_total") {
			if (it is IntTag)
				workingTicksTotal = it.asInt
		}
	}

	private val consumers = ArrayList<LazyOptional<IEnergyStorage>>()

	fun checkSurroundings() {
		if (!valid)
			return

		val known = consumers.clone() as ArrayList<*>
		consumers.clear()
		val level = level ?: return

		for (direction in Direction.values()) {
			// нельзя выталкивать энергию через перед
			if (direction == blockState.getValue(RotatableMatteryBlock.FACING))
				continue

			val resolver = level.getBlockEntity(blockPos + direction)?.getEnergySided(-direction)

			resolver?.ifPresentK {
				if (!known.contains(resolver)) {
					val ref = WeakReference(this)

					resolver.addListener {
						if (!SERVER_IS_DYING)
							ref.get()?.checkSurroundings()
					}
				}

				consumers.add(resolver)
			}
		}
	}

	val container = MatteryContainer(this::setChangedLight, 3)

	val itemHandler = container.handler(fun (slot, stack): Boolean {
		if (slot == SLOT_INPUT)
			return ForgeHooks.getBurnTime(stack, null) > 0

		if (slot == SLOT_RESIDUE)
			return false

		return stack.getCapability(ForgeCapabilities.ENERGY).isPresent
	}, fun (slot, _, stack): Boolean {
		if (slot == SLOT_RESIDUE) return true

		return slot == SLOT_BATTERY &&
				(!stack.getCapability(ForgeCapabilities.ENERGY).isPresent || stack.getCapability(ForgeCapabilities.ENERGY).resolve().get().receiveEnergy(Int.MAX_VALUE, true) <= 0)
	})

	override fun setBlockState(p_155251_: BlockState) {
		super.setBlockState(p_155251_)

		if (valid)
			tickOnceServer(this::checkSurroundings)
	}

	var workingTicks = 0
		private set

	var workingTicksTotal = 0
		private set

	private var check = true

	private fun workWithPower(it: IEnergyStorage) {
		val extracted = energy.extractEnergyInner(THROUGHPUT, true)
		val received = it.receiveEnergy(extracted, false)

		if (!received.isZero) {
			energy.extractEnergyInner(received, false)
		}
	}

	fun tick() {
		if (workingTicks > 0 && !isBlockedByRedstone) {
			workingTicks--
			energy.receiveEnergyInner(GENERATION_SPEED, false)

			if (workingTicks == 0) {
				workingTicksTotal = 0
				check = true
			}

			if (blockState.getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.WORKING) {
				level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
			}
		} else if (blockState.getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.IDLE) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
		}

		if (workingTicks == 0 && !isBlockedByRedstone && check) {
			if (!container[SLOT_INPUT].isEmpty) {
				val ticks = ForgeHooks.getBurnTime(container[SLOT_INPUT], null)
				val residue = container[SLOT_INPUT].item.getCraftingRemainingItem(container[SLOT_INPUT].copy().also { it.count = 1 })
				val canPutResidue = residue.isEmpty || container[SLOT_RESIDUE].isEmpty || ItemStack.isSameItemSameTags(container[SLOT_RESIDUE], residue) && container[SLOT_RESIDUE].count < container[2].maxStackSize

				if (canPutResidue && ticks >= 4 && (energy.batteryLevel < ImpreciseFraction.ONE || GENERATION_SPEED * (ticks / 4) + energy.batteryLevel <= energy.maxBatteryLevel)) {
					workingTicksTotal = ticks / 4
					workingTicks = ticks / 4
					container[SLOT_INPUT].shrink(1)

					if (!residue.isEmpty) {
						if (container[SLOT_RESIDUE].isEmpty) {
							container[SLOT_RESIDUE] = residue
						} else {
							container[SLOT_RESIDUE].count++
						}
					}

					container.setChanged(SLOT_INPUT)
				}
			}

			check = false
		}

		if (energy.batteryLevel.isZero) return

		val item = container[SLOT_BATTERY]

		if (!item.isEmpty) {
			item.energy?.let(this::workWithPower)
			if (energy.batteryLevel.isZero) return
		}

		for (consumer in consumers) {
			consumer.ifPresent(this::workWithPower)
		}
	}

	companion object {
		private val THROUGHPUT = ImpreciseFraction(160)
		private const val THROUGHPUT_INT = 160
		private val GENERATION_SPEED = ImpreciseFraction(40)
		private val MAX_ENERGY = ImpreciseFraction(24_000)
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.chemical_generator")

		const val SLOT_INPUT = 0
		const val SLOT_BATTERY = 1
		const val SLOT_RESIDUE = 2
	}
}
