package ru.dbotthepony.mc.otm.block.entity.storage

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import net.minecraftforge.items.IItemHandler
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.CableBlock
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.block.entity.MatteryPoweredBlockEntity
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.GraphNodeListener
import ru.dbotthepony.mc.otm.graph.storage.BasicStorageGraphNode
import ru.dbotthepony.mc.otm.graph.storage.StorageNetworkGraph
import ru.dbotthepony.mc.otm.menu.StorageExporterMenu
import ru.dbotthepony.mc.otm.menu.StorageImporterMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.storage.*
import java.math.BigInteger
import java.util.*
import java.util.stream.Stream
import kotlin.collections.HashSet

abstract class AbstractStorageImportExport<T>(
	blockType: BlockEntityType<*>,
	blockPos: BlockPos,
	blockState: BlockState,
	maxBatteryLevel: ImpreciseFraction = MAX_POWER,
	maxInput: ImpreciseFraction? = BlockEnergyStorageImpl.DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : MatteryPoweredBlockEntity(blockType, blockPos, blockState) {
	final override val energy = WorkerEnergyStorage(this, maxBatteryLevel, maxInput, maxOutput)

	val cell: BasicStorageGraphNode = object : BasicStorageGraphNode(energy), GraphNodeListener {
		override fun onNeighbour(node: Graph6Node<*>, direction: Direction) {
			val newState = this@AbstractStorageImportExport.blockState.setValue(CableBlock.MAPPING_CONNECTION_PROP[direction.ordinal], true)

			if (newState !== this@AbstractStorageImportExport.blockState && !SERVER_IS_DYING)
				level?.setBlock(blockPos, newState, Block.UPDATE_CLIENTS)
		}

		override fun onUnNeighbour(node: Graph6Node<*>, direction: Direction) {
			val newState = this@AbstractStorageImportExport.blockState.setValue(CableBlock.MAPPING_CONNECTION_PROP[direction.ordinal], false)

			if (newState !== this@AbstractStorageImportExport.blockState && !SERVER_IS_DYING)
				level?.setBlock(blockPos, newState, Block.UPDATE_CLIENTS)
		}
	}

	private var valid = true

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (valid && cap === MatteryCapability.STORAGE_NODE && side != blockState.getValue(RotatableMatteryBlock.FACING_FULL)) {
			cell.get().cast()
		} else super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		cell.invalidate()
		valid = false
	}

	override fun reviveCaps() {
		super.reviveCaps()
		cell.revive()
		valid = true
	}

	override fun setRemoved() {
		super.setRemoved()
		cell.destroy(level)
		valid = false
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel) {
			StorageNetworkGraph.discoverFull(this, cell.storageNode)
		}

		tickOnceServer(this::checkSurroundings)
	}

	protected var target: LazyOptional<T> = LazyOptional.empty()
	protected abstract val targetCapability: Capability<T>

	fun checkSurroundings() {
		target = getAndBind(
			target,
			level?.getBlockEntity(blockPos + blockState.getValue(RotatableMatteryBlock.FACING_FULL).normal),
			targetCapability,
			-blockState.getValue(RotatableMatteryBlock.FACING_FULL),
		) { tickOnceServer(this::checkSurroundings) }
	}

	companion object {
		val MAX_POWER = ImpreciseFraction(10_000)
	}
}

class StorageImporterBlockEntity(blockPos: BlockPos, blockState: BlockState)
	: AbstractStorageImportExport<IItemHandler>(MBlockEntities.STORAGE_IMPORTER, blockPos, blockState),
	IItemHandler {
	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	val filter = ItemFilter(MAX_FILTERS) { _, _, _ ->
		setChangedLight()
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return StorageImporterMenu(containerID, inventory, this)
	}

	private var lastSlot = 0
	private var nextTick = INTERVAL

	private val enoughEnergy get() = energy.batteryLevel >= ITEM_STORAGE.energyPerOperation

	override val targetCapability: Capability<IItemHandler>
		get() = CapabilityItemHandler.ITEM_HANDLER_CAPABILITY

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		nbt["filter"] = filter.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("filter", CompoundTag::class.java, filter::deserializeNBT)
	}

	private var valid = true
	private var resolverItemHandler = LazyOptional.of<IItemHandler> { this }

	override fun invalidateCaps() {
		super.invalidateCaps()
		resolverItemHandler.invalidate()
		valid = false
	}

	override fun reviveCaps() {
		super.reviveCaps()
		resolverItemHandler = LazyOptional.of { this }
		valid = true
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid && cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY && side == blockState.getValue(RotatableMatteryBlock.FACING_FULL)) {
			return resolverItemHandler.cast()
		}

		return super.getCapability(cap, side)
	}

	override fun getSlots(): Int {
		return 1
	}

	override fun getStackInSlot(slot: Int): ItemStack {
		return ItemStack.EMPTY
	}

	override fun insertItem(slot: Int, stack: ItemStack, simulate: Boolean): ItemStack {
		if (!filter.match(stack))
			return stack

		val view = cell.storageGraph?.getVirtualComponent(ITEM_STORAGE) ?: return stack
		val maxMove = energy.extractStepInnerBi(ITEM_STORAGE.energyPerOperation, stack.count, true)

		if (maxMove == BigInteger.ZERO)
			return stack

		val leftover = view.insertStack(ItemStackWrapper(stack).also { it.count = maxMove }, simulate)

		if (simulate)
			return leftover.stack

		energy.extractStepInner(ITEM_STORAGE.energyPerOperation, maxMove - leftover.count, false)
		return leftover.stack
	}

	override fun extractItem(slot: Int, amount: Int, simulate: Boolean): ItemStack {
		return ItemStack.EMPTY
	}

	override fun getSlotLimit(slot: Int): Int {
		return Int.MAX_VALUE
	}

	override fun isItemValid(slot: Int, stack: ItemStack): Boolean {
		return filter.match(stack)
	}

	fun tick() {
		batteryChargeLoop()
		cell.tickEnergyDemanding()

		nextTick--

		if (nextTick <= 0 && target.isPresent && enoughEnergy) {
			val graph = cell.storageGraph ?: return
			val items = graph.getVirtualComponent(ITEM_STORAGE)

			val resolved = target.orThrow()

			if (lastSlot >= resolved.slots) {
				lastSlot = 0
			}

			val maxMove = energy.extractStepInner(ITEM_STORAGE.energyPerOperation, MAX_MOVE_PER_OPERATION, true)
			var extracted = resolved.extractItem(lastSlot, maxMove, true)

			if (extracted.isEmpty || !filter.match(extracted)) {
				lastSlot++
			} else {
				val leftOver = items.insertStack(ItemStackWrapper(extracted), true)

				if (leftOver.count.toInt() != extracted.count) {
					extracted = resolved.extractItem(lastSlot, extracted.count - leftOver.count.toInt(), false)
					energy.extractStepInner(ITEM_STORAGE.energyPerOperation, extracted.count, false)
					items.insertStack(ItemStackWrapper(extracted), false)
				} else {
					nextTick += INTERVAL * 4
				}
			}
		}

		if (nextTick <= 0) {
			nextTick = INTERVAL
		}
	}

	companion object {
		const val MAX_MOVE_PER_OPERATION = 4
		private val MACHINE_NAME = TranslatableComponent("block.${OverdriveThatMatters.MOD_ID}.${MNames.STORAGE_IMPORTER}")
		private const val INTERVAL = 5
		const val MAX_FILTERS = 6 * 3
	}
}

class StorageExporterBlockEntity(blockPos: BlockPos, blockState: BlockState) :
	AbstractStorageImportExport<IItemHandler>(MBlockEntities.STORAGE_EXPORTER, blockPos, blockState),
	IStorageEventConsumer<ItemStackWrapper> {
	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return StorageExporterMenu(containerID, inventory, this)
	}

	private val relevantTuples = HashSet<UUID>()

	override val storageType: StorageStackType<ItemStackWrapper>
		get() = ITEM_STORAGE

	init {
		cell.addStorageComponent(this)
	}

	override fun addStack(stack: ItemStackWrapper, id: UUID, provider: IStorageProvider<ItemStackWrapper>) {
		if (!filter.match(stack.item)) {
			return
		}

		relevantTuples.add(id)
	}

	override fun changeStack(stack: ItemStackWrapper, id: UUID, oldCount: BigInteger) {
		// no-op
	}

	override fun removeStack(stack: ItemStackWrapper, id: UUID) {
		relevantTuples.remove(id)
	}

	val filter = ItemFilter(MAX_FILTERS) { _, _, _ ->
		relevantTuples.clear()

		val component = cell.storageGraph?.getVirtualComponent(ITEM_STORAGE) ?: return@ItemFilter

		for (tuple in component.stacks) {
			addStack(tuple, component)
		}

		lastSlot = 0

		setChangedLight()
	}.also { it.isWhitelist = true }

	private var lastSlot = 0
	private var nextTick = INTERVAL
	private val enoughEnergy get() = energy.batteryLevel >= ITEM_STORAGE.energyPerOperation

	override val targetCapability: Capability<IItemHandler>
		get() = CapabilityItemHandler.ITEM_HANDLER_CAPABILITY

	private val exportStacks: Stream<Pair<UUID, ItemStackWrapper>>
		get() {
			val view = cell.storageGraph?.getVirtualComponent(ITEM_STORAGE) ?: return Stream.empty()
			return relevantTuples.stream().map { it to view[it] }
		}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		nbt["filter"] = filter.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("filter", CompoundTag::class.java, filter::deserializeNBT)
	}

	fun tick() {
		batteryChargeLoop()
		cell.tickEnergyDemanding()

		nextTick--

		if (nextTick <= 0 && target.isPresent && enoughEnergy) {
			val graph = cell.storageGraph ?: return
			val items = graph.getVirtualComponent(ITEM_STORAGE)

			val resolved = target.orThrow()

			if (lastSlot >= resolved.slots) {
				lastSlot = 0
			}

			var hit = false

			for (stack in exportStacks) {
				if (!resolved.isItemValid(lastSlot, stack.second.item)) {
					continue
				}

				val exportAmountA = items.extractStack(stack.first, stack.second.count.coerceAtMost(
					MAX_MOVE_PER_OPERATION
				), true).count

				if (exportAmountA == BigInteger.ZERO) {
					continue
				}

				var exportAmount = energy.extractStepInner(ITEM_STORAGE.energyPerOperation, exportAmountA, true)

				if (exportAmount == 0) {
					break
				}

				val leftover = resolved.insertItem(lastSlot, stack.second.stack.also { it.count = exportAmount }, true)

				if (leftover.count != exportAmount) {
					hit = true
					exportAmount = items.extractStack(stack.first, (exportAmount - leftover.count).toBigInteger(), false).count.toInt()
					resolved.insertItem(lastSlot, stack.second.stack.also { it.count = exportAmount }, false)
					energy.extractStepInner(ITEM_STORAGE.energyPerOperation, exportAmount, false)
					break
				}
			}

			if (!hit) {
				lastSlot++
			}
		}

		if (nextTick <= 0) {
			nextTick = INTERVAL
		}
	}

	companion object {
		val MAX_MOVE_PER_OPERATION: BigInteger = BigInteger.valueOf(4L)
		private val MACHINE_NAME = TranslatableComponent("block.${OverdriveThatMatters.MOD_ID}.${MNames.STORAGE_EXPORTER}")
		private const val INTERVAL = 5
		const val MAX_FILTERS = 6 * 3
	}
}
