package ru.dbotthepony.mc.otm.client.screen

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectFunction
import net.minecraft.ChatFormatting
import net.minecraft.client.gui.Font
import net.minecraft.client.gui.screens.inventory.AbstractContainerScreen
import net.minecraft.client.renderer.entity.ItemRenderer
import net.minecraft.network.chat.Component
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.inventory.Slot
import net.minecraftforge.client.event.ContainerScreenEvent.Render.Background
import net.minecraftforge.client.event.ContainerScreenEvent.Render.Foreground
import net.minecraftforge.common.MinecraftForge
import org.lwjgl.opengl.GL11
import org.lwjgl.opengl.GL13
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.screen.panels.*
import ru.dbotthepony.mc.otm.menu.MatteryMenu

/**
 * This class encapsulate most of logic for handling EditablePanel and it's children.
 *
 * Widgets such as slots are not drawn/interacted with using MatteryScreen,
 * but are handled by appropriate panels
 *
 * @param <T> type of menu, which extends MatteryMenu
 */
@Suppress("LeakingThis")
abstract class MatteryScreen<T : MatteryMenu>(menu: T, inventory: Inventory, title: Component) : AbstractContainerScreen<T>(menu, inventory, title) {
	constructor(menu: T, title: Component) : this(menu, menu.inventory, title)

	protected val panels = ArrayDeque<EditablePanel>()

	var inventoryFrame: FramePanel? = null
	var mainFrame: FramePanel? = null
	private var madeMainFrame = false

	var itemRenderer: ItemRenderer
		get() = itemRenderer
		set(itemRenderer) { super.itemRenderer = itemRenderer }
	var font: Font
		get() = font
		set(font) { super.font = font }
	var hoveredSlot: Slot?
		get() = hoveredSlot
		set(value) { super.hoveredSlot = value }
	val quickCraftSlots: MutableSet<Slot> get() = quickCraftSlots
	val quickCraftingType get() = quickCraftingType
	val isQuickCrafting get() = isQuickCrafting

	private val inventorySlotsRows = Int2ObjectAVLTreeMap<EditablePanel>()

	init {
		if (menu.playerInventorySlots.isNotEmpty() && menu.autoCreateInventoryFrame) {
			if (menu.playerExoSuitSlots.isEmpty()) {
				inventoryFrame = FramePanel(this, null, 0f, 0f, INVENTORY_FRAME_WIDTH, INVENTORY_FRAME_HEIGHT, inventory.displayName).also(this::addPanel)

				val hotbarStrip = EditablePanel(this, inventoryFrame, height = AbstractSlotPanel.SIZE)
				hotbarStrip.dock = Dock.BOTTOM

				for (slot in menu.playerHotbarSlots) {
					SlotPanel(this, hotbarStrip, slot).also { it.dock = Dock.LEFT }
				}

				val canvas = EditablePanel(this, inventoryFrame)

				canvas.dock = Dock.FILL

				for (i in 0 .. 2) {
					getInventorySlotsRow(i).also {
						it.parent = inventoryFrame
						it.y = i * AbstractSlotPanel.SIZE
						it.parent = canvas
					}
				}
			} else {
				inventoryFrame = FramePanel(this, null, 0f, 0f, INVENTORY_FRAME_WIDTH_EXTENDED, INVENTORY_FRAME_HEIGHT, inventory.displayName).also(this::addPanel)

				var slotListCanvas: EditablePanel? = null

				val scrollbar = DiscreteScrollBarPanel(this, inventoryFrame, { ((menu.playerCombinedInventorySlots.size - 27) + 8) / 9 }, {
						_, old, new ->

					for (i in old .. old + 2) {
						getInventorySlotsRow(i).visible = false
					}

					for (i in new .. new + 2) {
						getInventorySlotsRow(i).also {
							it.visible = true
							it.parent = slotListCanvas
							it.y = (i - new) * AbstractSlotPanel.SIZE
						}
					}

					menu.ply.matteryPlayer?.exoSuitMenu?.lastScroll = new
				})

				slotListCanvas = object : EditablePanel(this@MatteryScreen, inventoryFrame, height = AbstractSlotPanel.SIZE * 3f) {
					override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
						scrollbar.mouseScrolledInner(x, y, scroll)
						return true
					}
				}

				slotListCanvas.dock = Dock.TOP

				val hotbarStrip = EditablePanel(this, inventoryFrame, height = AbstractSlotPanel.SIZE)
				hotbarStrip.dock = Dock.BOTTOM

				for (slot in menu.playerHotbarSlots) {
					SlotPanel(this, hotbarStrip, slot).also { it.dock = Dock.LEFT }
				}

				scrollbar.parent = slotListCanvas
				scrollbar.dock = Dock.RIGHT

				scrollbar.scroll = menu.ply.matteryPlayer?.exoSuitMenu?.lastScroll ?: 0

				for (i in scrollbar.scroll .. scrollbar.scroll + 2) {
					getInventorySlotsRow(i).also {
						it.parent = slotListCanvas
						it.y = (i - scrollbar.scroll) * AbstractSlotPanel.SIZE
					}
				}
			}
		}
	}

	protected fun getInventorySlotsRow(index: Int): EditablePanel {
		return inventorySlotsRows.computeIfAbsent(index, Int2ObjectFunction {
			val canvas = object : EditablePanel(this@MatteryScreen, null, width = AbstractSlotPanel.SIZE * 9f, height = AbstractSlotPanel.SIZE) {
				override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
					return false
				}
			}

			val offset = it * 9

			if (menu.playerCombinedInventorySlots.size <= offset) {
				return@Int2ObjectFunction canvas
			}

			for (i in 0 .. (8).coerceAtMost(menu.playerCombinedInventorySlots.size - offset - 1)) {
				val slot = object : SlotPanel<Slot>(this@MatteryScreen, canvas, menu.playerCombinedInventorySlots[offset + i]) {
					override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
						return false
					}
				}

				slot.dock = Dock.LEFT
			}

			return@Int2ObjectFunction canvas
		})
	}

	override fun init() {
		super.init()

		if (!madeMainFrame) {
			madeMainFrame = true
			mainFrame = makeMainFrame()?.also(this::addPanel)
			movePanels()
		}

		movePanels()
	}

	fun addPanel(panel: EditablePanel): Boolean {
		if (!panels.contains(panel)) {
			panels.addFirst(panel)
			return true
		}

		return false
	}

	fun removePanel(panel: EditablePanel): Boolean {
		val indexOf = panels.indexOf(panel)

		if (indexOf != -1) {
			panels.removeAt(indexOf)
			return true
		}

		return false
	}

	/**
	 * Moves panel at top of all other panels
	 *
	 * @param panel panel to be pushed up
	 */
	fun popup(panel: EditablePanel) {
		val indexOf = panels.indexOf(panel)

		require(indexOf != -1) { "No such panel $panel" }

		if (indexOf == 0) {
			return
		}

		panels.removeAt(indexOf)
		panels.addFirst(panel)
	}

	/**
	 * Safely to be overriden. By default, creates well-known dimensions window
	 *
	 * can be freely overridden by child classes
	 *
	 * @return FramePanel created, or null
	 */
	protected open fun makeMainFrame(): FramePanel? {
		return FramePanel(this, null, 0f, 0f, DEFAULT_FRAME_WIDTH, DEFAULT_FRAME_HEIGHT, getTitle())
	}

	public override fun recalculateQuickCraftRemaining() {
		super.recalculateQuickCraftRemaining()
	}

	fun movePanels() {
		var top = height / 2
		val mainFrame = mainFrame
		val inventoryFrame = inventoryFrame

		if (mainFrame != null) {
			top -= (mainFrame.height / 2).toInt()
		}

		if (inventoryFrame != null) {
			top -= (inventoryFrame.height / 2).toInt()
		}

		if (mainFrame != null) {
			mainFrame.setPos(width / 2 - mainFrame.width / 2, top.toFloat())
			top += mainFrame.height.toInt()
		}

		if (inventoryFrame != null) {
			inventoryFrame.setPos(width / 2 - inventoryFrame.width / 2, top.toFloat())
			top += inventoryFrame.height.toInt()
		}
	}

	override fun mouseClicked(x: Double, y: Double, button: Int): Boolean {
		var click = false

		for (panel in panels) {
			if (click || !panel.mouseClickedChecked(x, y, button)) {
				panel.killFocusForEverythingExceptInner()
			} else {
				if (returnSlot != null) {
					super.mouseClicked(x, y, button)
					returnSlot = null
				}

				click = true
			}
		}

		return true
	}

	override fun mouseDragged(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		for (panel in panels) {
			if (panel.mouseDraggedChecked(x, y, button, xDelta, yDelta)) {
				if (returnSlot != null) {
					super.mouseDragged(x, y, button, xDelta, yDelta)
					returnSlot = null
				}

				return true
			}
		}

		return true
	}

	override fun mouseReleased(p_97812_: Double, p_97813_: Double, p_97814_: Int): Boolean {
		for (panel in panels) {
			if (panel.mouseReleasedChecked(p_97812_, p_97813_, p_97814_)) {
				if (returnSlot != null) {
					super.mouseReleased(p_97812_, p_97813_, p_97814_)
					returnSlot = null
				}

				return true
			}
		}

		return super.mouseReleased(p_97812_, p_97813_, p_97814_)
	}

	override fun mouseScrolled(p_94686_: Double, p_94687_: Double, p_94688_: Double): Boolean {
		for (panel in panels) {
			if (panel.mouseScrolledChecked(p_94686_, p_94687_, p_94688_)) {
				return true
			}
		}

		return true
	}

	override fun keyReleased(p_94715_: Int, p_94716_: Int, p_94717_: Int): Boolean {
		for (panel in panels) {
			if (panel.keyReleased(p_94715_, p_94716_, p_94717_)) {
				return true
			}
		}

		return true
	}

	override fun charTyped(p_94683_: Char, p_94684_: Int): Boolean {
		for (panel in panels) {
			if (panel.charTyped(p_94683_, p_94684_)) {
				return true
			}
		}

		return true
	}

	override fun keyPressed(key: Int, scancode: Int, mods: Int): Boolean {
		for (panel in panels) {
			if (panel.keyPressed(key, scancode, mods)) {
				if (returnSlot != null) {
					super.keyPressed(key, scancode, mods)
					returnSlot = null
				}

				return true
			}
		}

		return super.keyPressed(key, scancode, mods)
	}

	override fun renderBg(p_97787_: PoseStack, p_97788_: Float, p_97789_: Int, p_97790_: Int) {}

	var returnSlot: Slot? = null

	override fun findSlot(x: Double, y: Double): Slot? {
		if (returnSlot != null) {
			return returnSlot
		}

		val xFloat = x.toFloat()
		val yFloat = y.toFloat()

		for (panel in panels) {
			val (status, slot) = panel.findSlot(xFloat, yFloat)

			if (status) {
				return slot
			}
		}

		return null
	}

	override fun render(poseStack: PoseStack, mouseX: Int, mouseY: Int, partialTick: Float) {
		val mouseXf = mouseX.toFloat()
		val mouseYf = mouseY.toFloat()

		// dark background
		this.renderBackground(poseStack)

		hoveredSlot = null
		var hovered = false

		for (panel in panels) {
			if (hovered) {
				panel.unsetHovered()
			} else if (panel.tickHover(mouseXf, mouseYf)) {
				hovered = true
			}
		}

		RenderSystem.defaultBlendFunc()
		RenderSystem.enableBlend()
		RenderSystem.enableDepthTest()
		RenderSystem.enableTexture()
		RenderSystem.activeTexture(GL13.GL_TEXTURE0)

		var depth = -900f

		for (i in panels.indices.reversed()) {
			val panel = panels[i]
			RenderSystem.depthFunc(GL11.GL_ALWAYS)
			panel.accumulatedDepth = depth
			RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
			depth = depth.coerceAtLeast(panel.render(poseStack, mouseXf, mouseYf, partialTick))
		}

		RenderSystem.depthFunc(GL11.GL_LESS)
		MinecraftForge.EVENT_BUS.post(Background(this, poseStack, mouseX, mouseY))
		RenderSystem.disableDepthTest()

		// Screen.super.render
		for (widget in renderables) {
			widget.render(poseStack, mouseX, mouseY, partialTick)
		}
		// /Screen.super.render

		RenderSystem.disableDepthTest()
		MinecraftForge.EVENT_BUS.post(Foreground(this, poseStack, mouseX, mouseY))

		var itemstack = if (draggingItem.isEmpty) menu.carried else draggingItem

		if (!itemstack.isEmpty) {
			val i2 = if (draggingItem.isEmpty) 8 else 16
			var overrideCount: String? = null

			if (isQuickCrafting && quickCraftSlots.size > 1) {
				itemstack = itemstack.copy()
				itemstack.count = quickCraftingRemainder

				if (itemstack.isEmpty) {
					overrideCount = ChatFormatting.YELLOW.toString() + "0"
				}
			}

			@Suppress("NULLABILITY_MISMATCH_BASED_ON_JAVA_ANNOTATIONS")
			renderFloatingItem(itemstack, mouseX - 8, mouseY - i2, overrideCount)
		}

		RenderSystem.disableDepthTest()

		// forge's hard limit tooltip Z position to 500 units
		// so if we have anything drawn above 500 units, lift it up
		if (depth > 500f) {
			poseStack.pushPose()
			poseStack.translate(0.0, 0.0, depth - 500.0)
		}

		if (menu.carried.isEmpty) {
			val hoveredSlot = hoveredSlot

			if (hoveredSlot != null && hoveredSlot.hasItem()) {
				this.renderTooltip(poseStack, hoveredSlot.item, mouseX, mouseY)
			} else {
				for (panel in panels) {
					if (panel.renderTooltips(poseStack, mouseXf, mouseYf, partialTick)) {
						break
					}
				}
			}
		}

		if (depth > 500f) {
			poseStack.popPose()
		}

		RenderSystem.enableDepthTest()
	}

	override fun containerTick() {
		for (panel in panels) {
			panel.tick()
		}
	}

	companion object {
		const val DEFAULT_FRAME_WIDTH = AbstractSlotPanel.SIZE * 9f + 16f
		const val DEFAULT_FRAME_HEIGHT = 100f

		const val INVENTORY_FRAME_WIDTH = DEFAULT_FRAME_WIDTH
		const val INVENTORY_FRAME_WIDTH_EXTENDED = DEFAULT_FRAME_WIDTH + ScrollBarConstants.WIDTH + 2f
		const val INVENTORY_FRAME_HEIGHT = 3f * AbstractSlotPanel.SIZE + AbstractSlotPanel.SIZE + 24f

		const val GAUGE_TOP_WITH_SLOT = 17f
		const val GAUGE_TOP_WITHOUT_SLOT = 26f
		const val SLOT_TOP_UNDER_GAUGE = 70f
		const val PROGRESS_ARROW_TOP = 42f
		const val PROGRESS_SLOT_TOP = 41f
		const val LEFT_MARGIN = 8f
	}
}