package ru.dbotthepony.mc.otm.client.screen.panels

import com.google.common.collect.ImmutableList
import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.Minecraft
import net.minecraft.client.gui.Font
import net.minecraft.client.gui.components.events.GuiEventListener
import net.minecraft.network.chat.Component
import net.minecraft.world.inventory.Slot
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.client.render.popScissorRect
import ru.dbotthepony.mc.otm.client.render.pushScissorRect
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import java.util.*
import kotlin.collections.ArrayList
import kotlin.math.max

@JvmRecord
data class ScreenPos(val x: Float, val y: Float)

@JvmRecord
data class DockProperty @JvmOverloads constructor(val left: Float = 0f, val top: Float = 0f, val right: Float = 0f, val bottom: Float = 0f) {
	val isEmpty get() = left == 0f && right == 0f && top == 0f && bottom == 0f

	companion object {
		val EMPTY = DockProperty()
	}
}

enum class Dock {
	NONE, LEFT, RIGHT, TOP, BOTTOM, FILL
}

enum class DockResizeMode(val changeWidth: Boolean, val changeHeight: Boolean) {
	ALL(true, true), NONE(false, false), WIDTH(true, false), HEIGHT(false, true)
}

interface ISlotPanel<S : Slot> {
	val slot: S
}

open class EditablePanel @JvmOverloads constructor(
	val screen: MatteryScreen<*>,
	parent: EditablePanel?,

	// относительно родителя
	x: Float = 0f,
	y: Float = 0f,

	width: Float = 10f,
	height: Float = 10f,
) : GuiEventListener {
	var parent: EditablePanel? = null
		set(value) {
			if (field === value)
				return

			field?.onUnParent(this)
			field = value
			layoutInvalidated = true
			value?.onParent(this)
		}

	var x: Float = 0f
		set(value) {
			if (field == value) return

			val old = field
			field = value
			// layoutInvalidated = true
			xPosUpdated(value, old)
			parent?.boundsInvalidated = true
		}

	var y: Float = 0f
		set(value) {
			if (field == value) return

			val old = field
			field = value
			// layoutInvalidated = true
			yPosUpdated(value, old)
			parent?.boundsInvalidated = true
		}

	var width: Float = 10f
		set(value) {
			if (field == value) return

			assert(value >= 0f) {"Invalid width $value"}

			val old = field
			field = value
			layoutInvalidated = true
			widthUpdated(value, old)
			parent?.boundsInvalidated = true
		}

	var height: Float = 10f
		set(value) {
			if (field == value) return

			assert(value >= 0f) {"Invalid height $value"}

			val old = field
			field = value
			layoutInvalidated = true
			heightUpdated(value, old)
			parent?.boundsInvalidated = true
		}

	/**
	 * Width to be utilized in docking code.
	 *
	 * Can only differ from [width] if [dockResize] is not [DockResizeMode.ALL] and not [DockResizeMode.WIDTH]
	 */
	var dockedWidth: Float = 0f
		private set

	/**
	 * Height to be utilized in docking code.
	 *
	 * Can only differ from [height] if [dockResize] is not [DockResizeMode.ALL] and not [DockResizeMode.HEIGHT]
	 */
	var dockedHeight: Float = 0f
		private set

	@JvmField
	protected val children = ArrayList<EditablePanel>()
	val childrenView: List<EditablePanel> = Collections.unmodifiableList(children)

	var layoutInvalidated = true
		private set
	var boundsInvalidated = true
		private set

	protected var visibleAsChildren: Boolean = true
		private set

	var visible: Boolean = true
		set(value) {
			if (field != value) {
				val old = isVisible()
				field = value
				val new = isVisible()

				if (old != new) {
					visibilityChanges(new, old)
					parent?.layoutInvalidated = true
					layoutInvalidated = true

					if (new) {
						killFocus()
					}
				}

				updateVisible()
			}
		}

	// Позволяет смещать потомков на эту координату
	var xOffset = 0f
	var yOffset = 0f

	// Обрезать отрисовку потомков?
	var scissor = false
	var dock = Dock.NONE
		set(value) {
			if (field != value) {
				field = value
				parent?.layoutInvalidated = true
			}
		}
	var dockResize = DockResizeMode.ALL
		set(value) {
			if (field != value) {
				field = value
				parent?.layoutInvalidated = true
			}
		}
	var dockMargin = DockProperty.EMPTY
		set(value) {
			if (field != value) {
				field = value
				parent?.layoutInvalidated = true
			}
		}

	var dockLeft: Float
		get() = dockMargin.left
		set(value) {
			if (value != dockMargin.left) {
				dockMargin = dockMargin.copy(left = value)
			}
		}

	var dockRight: Float
		get() = dockMargin.right
		set(value) {
			if (value != dockMargin.right) {
				dockMargin = dockMargin.copy(right = value)
			}
		}

	var dockTop: Float
		get() = dockMargin.top
		set(value) {
			if (value != dockMargin.top) {
				dockMargin = dockMargin.copy(top = value)
			}
		}

	var dockBottom: Float
		get() = dockMargin.bottom
		set(value) {
			if (value != dockMargin.bottom) {
				dockMargin = dockMargin.copy(bottom = value)
			}
		}

	var dockPadding = DockProperty.EMPTY
		set(value) {
			if (field != value) {
				field = value
				parent?.layoutInvalidated = true
			}
		}

	var acceptMouseInput = true
	var acceptKeyboardInput = true
	var trapMouseInput = false

	var tooltip: Component? = null
		set(value) {
			if (value != null) {
				tooltipList = null
			}

			field = value
		}

	var tooltipList: List<Component>? = null
		set(value) {
			if (value != null) {
				tooltip = null
			}

			field = value
		}

	@JvmOverloads
	fun setDockMargin(left: Float = 0f, top: Float = 0f, right: Float = 0f, bottom: Float = 0f) {
		dockMargin = DockProperty(left, top, right, bottom)
	}

	@JvmOverloads
	fun setDockPadding(left: Float = 0f, top: Float = 0f, right: Float = 0f, bottom: Float = 0f) {
		dockPadding = DockProperty(left, top, right, bottom)
	}

	// абсолютное начало координат этой панели и потомков
	// негативно если потомок выходит за границы
	var boundingX = 0f
		private set
	var boundingY = 0f
		private set
	var boundingWidth = 0f
		private set
	var boundingHeight = 0f
		private set

	// Абсолютная координата, обновляется на отрисовке
	var absoluteX = 0f
		private set
	var absoluteY = 0f
		private set

	var isHovered = false
		private set(value) {
			if (value == field) {
				return
			}

			field = value

			if (value) {
				onHovered()
			} else {
				onUnHovered()
			}
		}

	fun unsetHovered() {
		isHovered = false

		for (child in children) {
			child.unsetHovered()
		}
	}

	protected open fun onHovered() {}
	protected open fun onUnHovered() {}

	var isFocused = false
		set(value) {
			if (field != value) {
				val old = field
				field = value

				onFocusChanged(old, value)
				findAbsoluteRoot().updateFocus()
			}
		}

	var autoKillFocus = false
		protected set

	private var focusedAsParent = false

	var accumulatedDepth = 0f

	fun accumulate3DHeight(value: Float) {
		accumulatedDepth += value
	}

	val font: Font get() = screen.font

	fun invalidateLayout() {
		layoutInvalidated = true
	}

	fun most3DHeight(): Float {
		var depth = accumulatedDepth

		for (child in children) {
			depth = max(depth, child.most3DHeight())
		}

		return depth
	}

	init {
		this.x = x
		this.y = y
		this.width = width
		this.height = height
		this.parent = parent
	}

	private fun onParent(child: EditablePanel) {
		if (children.contains(child)) throw IllegalStateException("Already containing $child")
		children.add(child)
		layoutInvalidated = true
		updateVisible()
	}

	private fun onUnParent(child: EditablePanel) {
		val indexOf = children.indexOf(child)
		if (indexOf == -1) throw IllegalStateException("Already not containing $child")
		children.removeAt(indexOf)
		layoutInvalidated = true
		updateVisible()
	}

	protected open fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {}
	protected open fun innerRenderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		return false
	}

	fun isVisible(): Boolean {
		return visible && visibleAsChildren
	}

	fun render(poseStack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float): Float {
		if (!isVisible()) {
			return 0f
		}

		var depth = 0f

		if (layoutInvalidated) {
			performLayout()
		} else if (boundsInvalidated) {
			updateBounds()
		}

		val parent = this.parent

		if (parent == null) {
			absoluteX = x
			absoluteY = y
		} else {
			if ((x < 0f || y < 0f || x + width > parent.width || y + height > parent.height) && parent.parent == null) {
				// no op - we updated ourselves in tickHover
			} else {
				isHovered = parent.isHovered &&
						mouseX >= absoluteX &&
						mouseX < absoluteX + width &&
						mouseY >= absoluteY &&
						mouseY < absoluteY + height
			}
		}

		val scissor = this.scissor

		if (scissor) {
			val scale = Minecraft.getInstance().window.guiScale

			pushScissorRect(
				(absoluteX * scale).toInt(),
				(absoluteY * scale).toInt(),
				(width * scale).toInt() + 1,
				(height * scale).toInt() + 1,
			)
		}

		poseStack.pushPose()
		poseStack.translate(absoluteX.toDouble(), absoluteY.toDouble(), accumulatedDepth.toDouble())
		innerRender(poseStack, mouseX, mouseY, partialTick)
		poseStack.popPose()

		for (child in children) {
			if (child.isVisible()) {
				child.accumulatedDepth = accumulatedDepth + 1
				child.absoluteX = absoluteX + child.x + xOffset
				child.absoluteY = absoluteY + child.y + yOffset

				RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
				depth = max(depth, child.render(poseStack, mouseX, mouseY, partialTick))
			}
		}

		if (scissor) {
			popScissorRect()
		}

		return depth
	}

	open fun tickHover(mouseX: Float, mouseY: Float): Boolean {
		if ((boundingHeight > height || boundingWidth > width || boundingX < 0 || boundingY < 0) && parent == null) {
			var hit = false

			for (child in children) {
				if (child.tickHover(mouseX, mouseY)) {
					hit = true
				}
			}

			isHovered = mouseX >= absoluteX &&
					mouseX <= absoluteX + width &&
					mouseY >= absoluteY &&
					mouseY <= absoluteY + height

			return hit
		}

		isHovered =
				mouseX >= absoluteX &&
				mouseX <= absoluteX + width &&
				mouseY >= absoluteY &&
				mouseY <= absoluteY + height

		return isHovered
	}

	fun findSlot(mouseX: Float, mouseY: Float): Pair<Boolean, Slot?> {
		if (!acceptMouseInput) {
			return (mouseX >= absoluteX &&
					mouseX <= absoluteX + width &&
					mouseY >= absoluteY &&
					mouseY <= absoluteY + height) to null
		}

		if (trapMouseInput && this is ISlotPanel<*>) {
			return true to this.slot
		}

		for (child in children) {
			val (status, slot) = child.findSlot(mouseX, mouseY)

			if (status) {
				return true to slot
			}
		}

		if (
			mouseX >= absoluteX &&
			mouseX <= absoluteX + width &&
			mouseY >= absoluteY &&
			mouseY <= absoluteY + height
		) {
			if (this is ISlotPanel<*>) {
				return true to this.slot
			}

			return true to null
		}

		return false to null
	}

	fun renderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		if (!isVisible()) {
			return false
		}

		for (child in children) {
			if (child.isVisible() && child.renderTooltips(stack, mouse_x, mouse_y, flag)) {
				return true
			}
		}

		if (innerRenderTooltips(stack, mouse_x, mouse_y, flag)) {
			return true
		}

		if (isHovered) {
			val tooltip = tooltip
			val tooltipList = tooltipList

			if (tooltip != null) {
				screen.renderComponentTooltip(
					stack,
					listOf(tooltip),
					mouse_x.toInt(),
					mouse_y.toInt(),
					screen.font
				)

				return true
			} else if (tooltipList != null) {
				screen.renderComponentTooltip(
					stack,
					tooltipList,
					mouse_x.toInt(),
					mouse_y.toInt(),
					screen.font
				)

				return true
			}
		}

		return false
	}

	@JvmOverloads
	fun localToScreen(_x: Float = 0f, _y: Float = 0f): ScreenPos {
		var parent = this.parent

		var x = this.x + _x
		var y = this.y + _y

		while (parent != null) {
			x += parent.x + parent.xOffset
			y += parent.y + parent.yOffset
			parent = parent.parent
		}

		return ScreenPos(x, y)
	}

	fun screenToLocal(_x: Float = 0f, _y: Float = 0f): ScreenPos {
		val pos = localToScreen()
		return ScreenPos(_x - pos.x, _y - pos.y)
	}

	fun screenToLocal(_x: Double = 0.0, _y: Double = 0.0): ScreenPos {
		val pos = localToScreen()
		return ScreenPos((_x - pos.x).toFloat(), (_y - pos.y).toFloat())
	}

	fun findRoot(): EditablePanel? {
		var parent: EditablePanel? = this.parent ?: return null

		while (true) {
			val parent2 = parent!!.parent

			if (parent2 != null) {
				parent = parent2
			} else {
				return parent
			}
		}
	}

	fun findAbsoluteRoot(): EditablePanel {
		return findRoot() ?: this
	}

	/**
	 * Performs layout instantly instead of doing so on next frame.
	 *
	 * In almost all cases you want to call [invalidateLayout], and not this.
	 */
	open fun performLayout() {
		layoutInvalidated = false

		var left = dockPadding.left
		var right = dockPadding.right
		var top = dockPadding.top
		var bottom = dockPadding.bottom

		for (child in children) {
			if (child.isVisible()) {
				when (child.dock) {
					Dock.NONE -> {}
					Dock.FILL -> {}

					Dock.LEFT -> {
						child.x = left + child.dockMargin.left
						child.y = top + child.dockMargin.top

						left += child.width + child.dockMargin.left + child.dockMargin.right

						child.dockedHeight = 1f.coerceAtLeast(height - top - bottom - child.dockMargin.top - child.dockMargin.bottom)

						if (child.dockResize.changeHeight)
							child.height = child.dockedHeight
						else if (child.height != child.dockedHeight)
							child.y += child.dockedHeight / 2f - child.height / 2f
					}

					Dock.RIGHT -> {
						child.x = width - right - child.width - child.dockMargin.right
						child.y = top + child.dockMargin.top

						right += child.width + child.dockMargin.left + child.dockMargin.right

						child.dockedHeight = 1f.coerceAtLeast(height - top - bottom - child.dockMargin.top - child.dockMargin.bottom)

						if (child.dockResize.changeHeight)
							child.height = child.dockedHeight
						else if (child.height != child.dockedHeight)
							child.y += child.dockedHeight / 2f - child.height / 2f
					}

					Dock.TOP -> {
						child.x = left + child.dockMargin.left
						child.y = top + child.dockMargin.top

						top += child.height + child.dockMargin.top + child.dockMargin.bottom

						child.dockedWidth = 1f.coerceAtLeast(width - left - right - child.dockMargin.left - child.dockMargin.right)

						if (child.dockResize.changeWidth)
							child.width = child.dockedWidth
						else if (child.width != child.dockedWidth)
							child.x += child.dockedWidth / 2f - child.width / 2f
					}

					Dock.BOTTOM -> {
						child.x = left + child.dockMargin.left
						child.y = height - bottom - child.height - child.dockMargin.bottom

						bottom += child.height + child.dockMargin.top + child.dockMargin.bottom

						child.dockedWidth = 1f.coerceAtLeast(width - left - right - child.dockMargin.left - child.dockMargin.right)

						if (child.dockResize.changeWidth)
							child.width = child.dockedWidth
						else if (child.width != child.dockedWidth)
							child.x += child.dockedWidth / 2f - child.width / 2f
					}
				}
			}
		}

		for (child in children) {
			if (child.isVisible() && child.dock == Dock.FILL) {
				val w = width - child.dockMargin.left - child.dockMargin.right - right - left
				val h = height - child.dockMargin.bottom - child.dockMargin.top - bottom - top

				if (w < 1f || h < 1f) {
					LOGGER.error("Unable to fit $child inside $this (FILL returned dimensions of $w $h)")
				}

				child.x = left + child.dockMargin.left
				child.y = top + child.dockMargin.top

				child.dockedWidth = w.coerceAtLeast(1f)
				child.dockedHeight = h.coerceAtLeast(1f)

				if (child.dockResize.changeHeight)
					child.height = child.dockedHeight
				else if (child.height != child.dockedHeight)
					child.y += child.dockedHeight / 2f - child.height / 2f

				if (child.dockResize.changeWidth)
					child.width = child.dockedWidth
				else if (child.width != child.dockedWidth)
					child.x += child.dockedWidth / 2f - child.width / 2f
			}
		}

		updateBounds()
	}

	fun updateBounds() {
		boundsInvalidated = false

		boundingX = 0f
		boundingY = 0f
		boundingWidth = 0f
		boundingHeight = 0f

		for (child in children) {
			if (child.isVisible()) {
				boundingX = boundingX.coerceAtMost(child.x)
				boundingY = boundingY.coerceAtMost(child.y)
				boundingWidth = boundingWidth.coerceAtLeast(child.x + child.width)
				boundingHeight = boundingHeight.coerceAtLeast(child.y + child.height)
			}
		}
	}

	/**
	 * Attempts to tightly fit dimensions to all children
	 *
	 * Performs layout if required
	 */
	open fun sizeToContents() {
		if (layoutInvalidated) {
			performLayout()
		}

		var width = 1f
		var height = 1f

		for (child in children) {
			if (child.isVisible()) {
				width = width.coerceAtLeast(child.x + child.width)
				height = height.coerceAtLeast(child.y + child.height)
			}
		}

		this.width = width
		this.height = height
	}

	companion object {
		private val LOGGER = LogManager.getLogger()!!
	}

	protected open fun visibilityChanges(new: Boolean, old: Boolean) {

	}

	private fun updateVisible() {
		for (child in children) {
			val old = child.isVisible()
			child.visibleAsChildren = isVisible()
			val new = child.isVisible()

			if (old != new) {
				child.visibilityChanges(new, old)
			}

			child.updateVisible()
		}
	}

	private fun killFocusInternal() {
		for (child in children) {
			child.killFocusInternal()
		}

		if (isFocused) {
			isFocused = false
		}
	}

	fun killFocus() {
		if (isEverFocused()) {
			killFocusInternal()
			findAbsoluteRoot().updateFocus()
		}
	}

	fun findHierarchicalFocus(): EditablePanel? {
		if (isFocused) {
			return this
		}

		for (child in children) {
			val focus = child.findHierarchicalFocus()

			if (focus != null) {
				return focus
			}
		}

		return null
	}

	fun hasHierarchicalFocus(): Boolean {
		return findHierarchicalFocus() != null
	}

	protected open fun onHierarchicalFocusChanged(new: Boolean, old: Boolean) {

	}

	protected open fun onFocusChanged(new: Boolean, old: Boolean) {

	}

	private fun updateFocus() {
		val old = focusedAsParent
		focusedAsParent = hasHierarchicalFocus()

		if (focusedAsParent != old) {
			onHierarchicalFocusChanged(focusedAsParent, old)
		}

		for (child in children) {
			child.updateFocus()
		}
	}

	fun isEverFocused(): Boolean {
		return isFocused || focusedAsParent
	}

	fun requestFocus() {
		if (isFocused) {
			return
		}

		if (focusedAsParent) {
			var child = findHierarchicalFocus()

			while (child != null) {
				child.isFocused = false
				child = findHierarchicalFocus()
			}
		}

		isFocused = true
	}

	protected open fun xPosUpdated(new: Float, old: Float) {}
	protected open fun yPosUpdated(new: Float, old: Float) {}
	protected open fun widthUpdated(new: Float, old: Float) {}
	protected open fun heightUpdated(new: Float, old: Float) {}

	fun setPos(x: Float, y: Float) {
		this.x = x
		this.y = y
	}

	fun setPos(x: Double, y: Double) {
		this.x = x.toFloat()
		this.y = y.toFloat()
	}

	fun setSize(w: Float, h: Float) {
		this.width = w
		this.height = h
	}

	fun setSize(w: Double, h: Double) {
		this.width = w.toFloat()
		this.height = h.toFloat()
	}

	fun setDimensions(x: Float, y: Float, w: Float, h: Float) {
		setPos(x, y)
		setSize(w, h)
	}

	fun setDimensions(x: Double, y: Double, w: Double, h: Double) {
		setPos(x, y)
		setSize(w, h)
	}

	fun fetchChildren(): List<EditablePanel> {
		return ImmutableList.copyOf(children)
	}

	fun getUndockedChildren(): List<EditablePanel> {
		var count = 0

		for (child in children) {
			if (child.dock == Dock.NONE) {
				count++
			}
		}

		val list = ArrayList<EditablePanel>(count)

		for (child in children) {
			if (child.dock == Dock.NONE) {
				list.add(child)
			}
		}

		return list
	}

	operator fun get(index: Int): EditablePanel? {
		if (index < 0 || index >= children.size) return null
		return children[index]
	}

	fun isTrappingMouseInput(): Boolean {
		if (trapMouseInput) {
			return true
		}

		for (child in children) {
			if (child.isTrappingMouseInput()) {
				return true
			}
		}

		return false
	}

	override fun mouseMoved(x: Double, y: Double) {

	}

	protected open fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		return true
	}

	fun withinBounds(x: Double, y: Double): Boolean {
		val pos: ScreenPos = localToScreen()
		return pos.x <= x && pos.x + width > x && pos.y <= y && pos.y + height > y
	}

	fun withinExtendedBounds(x: Double, y: Double): Boolean {
		val pos: ScreenPos = localToScreen(boundingX, boundingY)
		return pos.x <= x && pos.x + boundingWidth > x && pos.y <= y && pos.y + boundingHeight > y
	}

	fun killFocusForEverythingExcept(except: EditablePanel) {
		for (child in children) {
			if (child !== except) {
				child.killFocusForEverythingExceptInner()
			}
		}
	}

	fun killFocusForEverythingExceptInner() {
		for (child in children) {
			child.killFocusForEverythingExceptInner()
		}

		if (autoKillFocus) {
			killFocus()
		}
	}

	final override fun mouseClicked(x: Double, y: Double, button: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (trapMouseInput) return mouseClickedInner(x, y, button)

		for (child in children) {
			if (child.mouseClickedChecked(x, y, button)) {
				killFocusForEverythingExcept(child)
				return true
			}
		}

		return mouseClickedInner(x, y, button)
	}

	fun mouseClickedChecked(x: Double, y: Double, button: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isTrappingMouseInput() || withinBounds(x, y)) {
			if (acceptMouseInput && parent == null) screen.popup(this)
			return mouseClicked(x, y, button)
		} else if (withinExtendedBounds(x, y)) {
			for (child in children) {
				if (child.mouseClickedChecked(x, y, button)) {
					killFocusForEverythingExcept(child)
					return true
				}
			}

			if (autoKillFocus) killFocus()
		} else if (autoKillFocus) {
			killFocus()
		}

		return false
	}

	protected open fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		return true
	}

	final override fun mouseReleased(x: Double, y: Double, button: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (trapMouseInput) return mouseReleasedInner(x, y, button)

		for (child in children) {
			if (child.mouseReleasedChecked(x, y, button)) {
				return true
			}
		}

		return mouseReleasedInner(x, y, button)
	}

	fun mouseReleasedChecked(x: Double, y: Double, button: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isTrappingMouseInput() || withinBounds(x, y)) {
			return mouseReleased(x, y, button)
		} else if (withinExtendedBounds(x, y)) {
			for (child in children) {
				if (child.mouseReleasedChecked(x, y, button)) {
					return true
				}
			}
		}

		return false
	}

	protected open fun mouseDraggedInner(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		return true
	}

	final override fun mouseDragged(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (trapMouseInput) return mouseDraggedInner(x, y, button, xDelta, yDelta)

		for (child in children) {
			if (child.mouseDraggedChecked(x, y, button, xDelta, yDelta)) {
				return true
			}
		}

		return mouseDraggedInner(x, y, button, xDelta, yDelta)
	}

	fun mouseDraggedChecked(x: Double, y: Double, button: Int, xDelta: Double, yDelta: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isTrappingMouseInput() || withinBounds(x, y)) {
			return mouseDragged(x, y, button, xDelta, yDelta)
		} else if (withinExtendedBounds(x, y)) {
			for (child in children) {
				if (child.mouseDraggedChecked(x, y, button, xDelta, yDelta)) {
					return true
				}
			}
		}

		return false
	}


	protected open fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
		return true
	}

	final override fun mouseScrolled(x: Double, y: Double, scroll: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (trapMouseInput) return mouseScrolledInner(x, y, scroll)

		for (child in children) {
			if (child.mouseScrolledChecked(x, y, scroll)) {
				return true
			}
		}

		return mouseScrolledInner(x, y, scroll)
	}

	fun mouseScrolledChecked(x: Double, y: Double, scroll: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isTrappingMouseInput() || withinBounds(x, y)) {
			return mouseScrolled(x, y, scroll)
		} else if (withinExtendedBounds(x, y)) {
			for (child in children) {
				if (child.mouseScrolledChecked(x, y, scroll)) {
					return true
				}
			}
		}

		return false
	}

	protected open fun keyPressedInternal(key: Int, scancode: Int, mods: Int): Boolean {
		return false
	}

	final override fun keyPressed(key: Int, scancode: Int, mods: Int): Boolean {
		if (!isVisible() || !acceptKeyboardInput) return false
		if (!focusedAsParent) return false

		if (isFocused) return keyPressedInternal(key, scancode, mods)

		for (child in children) {
			if (child.keyPressed(key, scancode, mods)) {
				return true
			}
		}

		return false
	}

	protected open fun keyReleasedInternal(key: Int, scancode: Int, mods: Int): Boolean {
		return false
	}

	final  override fun keyReleased(key: Int, scancode: Int, mods: Int): Boolean {
		if (!isVisible() || !acceptKeyboardInput) return false
		if (!focusedAsParent) return false

		if (isFocused) return keyReleasedInternal(key, scancode, mods)

		for (child in children) {
			if (child.keyReleased(key, scancode, mods)) {
				return true
			}
		}

		return false
	}

	protected open fun charTypedInternal(codepoint: Char, mods: Int): Boolean {
		return false
	}

	final override fun charTyped(codepoint: Char, mods: Int): Boolean {
		if (!isVisible() || !acceptKeyboardInput) return false
		if (!focusedAsParent) return false

		if (isFocused) return charTypedInternal(codepoint, mods)

		for (child in children) {
			if (child.charTyped(codepoint, mods)) {
				return true
			}
		}

		return true
	}

	override fun changeFocus(state: Boolean): Boolean {
		return if (!isVisible()) false else super.changeFocus(state)
	}

	final override fun isMouseOver(x: Double, y: Double): Boolean { // called to check whenever we are hovering at this
		if (!isVisible() || !acceptMouseInput) return false
		if (isTrappingMouseInput()) return true

		val pos = localToScreen()
		return x >= pos.x && x <= pos.x + width && y >= pos.y && y + height <= pos.y
	}

	protected var tick = 0

	open fun tick() {
		tick++

		for (child in children) {
			child.tick()
		}
	}


	var isRemoved = false
		private set

	protected open fun onRemoved() {}

	fun remove() {
		if (isRemoved) {
			return
		}

		killFocus()
		screen.removePanel(this)

		for (child in children) {
			child.remove()
		}

		onRemoved()
		isRemoved = true
	}

	fun asGrid(): FlexGridPanel {
		val grid = FlexGridPanel(screen, parent, x, y, width, height)
		parent = grid
		grid.dock = dock
		dock = Dock.NONE
		grid.dockPadding = dockPadding
		grid.dockMargin = dockMargin
		setDockPadding(0f, 0f, 0f, 0f)
		setDockMargin(0f, 0f, 0f, 0f)
		return grid
	}
}
