package ru.dbotthepony.mc.otm.item

import com.google.common.collect.ImmutableMultimap
import com.google.common.collect.Multimap
import net.minecraft.ChatFormatting
import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.entity.EquipmentSlot
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.ai.attributes.Attribute
import net.minecraft.world.entity.ai.attributes.AttributeModifier
import net.minecraft.world.entity.ai.attributes.Attributes
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Material
import net.minecraftforge.common.capabilities.ICapabilityProvider
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.registry.EMPDamageSource

class EnergySwordItem : Item(Properties().stacksTo(1).rarity(Rarity.RARE).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
	val chargedAttributes: Multimap<Attribute, AttributeModifier>
	val dischargedAttributes: Multimap<Attribute, AttributeModifier>

	init {
		var builder = ImmutableMultimap.builder<Attribute, AttributeModifier>()

		builder.put(Attributes.ATTACK_DAMAGE, AttributeModifier(BASE_ATTACK_DAMAGE_UUID, "Weapon modifier", 13.0, AttributeModifier.Operation.ADDITION))
		builder.put(Attributes.ATTACK_SPEED, AttributeModifier(BASE_ATTACK_SPEED_UUID, "Weapon modifier", -2.4, AttributeModifier.Operation.ADDITION))

		chargedAttributes = builder.build()

		builder = ImmutableMultimap.builder()

		builder.put(Attributes.ATTACK_DAMAGE, AttributeModifier(BASE_ATTACK_DAMAGE_UUID, "Weapon modifier", 3.5, AttributeModifier.Operation.ADDITION))
		builder.put(Attributes.ATTACK_SPEED, AttributeModifier(BASE_ATTACK_SPEED_UUID, "Weapon modifier", -2.4, AttributeModifier.Operation.ADDITION))

		dischargedAttributes = builder.build()
	}

	override fun getDestroySpeed(itemStack: ItemStack, blockState: BlockState): Float {
		val energy = itemStack.getCapability(MatteryCapability.ENERGY).orNull() ?: return 1f

		if (blockState.`is`(Blocks.COBWEB)) {
			if (energy.batteryLevel < COBWEB_POWER_COST) {
				return 2f
			} else {
				return 25f
			}
		}

		return when (blockState.material) {
			Material.PLANT, Material.REPLACEABLE_PLANT, Material.VEGETABLE -> if (energy.batteryLevel < PLANT_POWER_COST) 1.5f else 4f
			Material.LEAVES -> if (energy.batteryLevel < PLANT_POWER_COST) 1.5f else 8f
			else -> 1f
		}
	}

	override fun canAttackBlock(p_41441_: BlockState, p_41442_: Level, p_41443_: BlockPos, p_41444_: Player): Boolean {
		return !p_41444_.isCreative
	}

	override fun hurtEnemy(itemStack: ItemStack, victim: LivingEntity, attacker: LivingEntity): Boolean {
		itemStack.getCapability(MatteryCapability.ENERGY).orNull()?.let {
			if (!it.extractEnergyInnerExact(ENERGY_PER_SWING, false).isZero) {
				victim.matteryPlayer?.let {
					it.androidEnergy.extractEnergyInner(ENERGY_ZAP, false)
					victim.hurt(EMPDamageSource(attacker), 8f)
				}
			}
		}

		return true
	}

	override fun isBarVisible(p_150899_: ItemStack): Boolean {
		return p_150899_.matteryEnergy != null
	}

	override fun getBarWidth(p_150900_: ItemStack): Int {
		return p_150900_.matteryEnergy?.getBarWidth() ?: super.getBarWidth(p_150900_)
	}

	override fun getBarColor(p_150901_: ItemStack): Int {
		return p_150901_.matteryEnergy?.getBarColor() ?: super.getBarColor(p_150901_)
	}

	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		p_41423_: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, p_41423_, p_41424_)

		itemStack.getCapability(MatteryCapability.ENERGY).ifPresentK {
			p_41423_.add(
				TranslatableComponent(
					"otm.item.power.normal.storage",
					it.batteryLevel.formatPower(),
					it.maxBatteryLevel.formatPower()
				).withStyle(ChatFormatting.GRAY)
			)
		}

		p_41423_.add(DESCRIPTION)
		p_41423_.add(DESCRIPTION2)
	}

	override fun mineBlock(
		itemStack: ItemStack,
		p_41417_: Level,
		blockState: BlockState,
		p_41419_: BlockPos,
		p_41420_: LivingEntity
	): Boolean {
		if (blockState.getDestroySpeed(p_41417_, p_41419_) != 0f) {
			val energy = itemStack.matteryEnergy

			when (blockState.material) {
				Material.PLANT, Material.REPLACEABLE_PLANT, Material.VEGETABLE, Material.LEAVES ->
					energy?.extractEnergyInnerExact(PLANT_POWER_COST, false)
			}

			if (blockState.`is`(Blocks.COBWEB)) {
				energy?.extractEnergyInnerExact(COBWEB_POWER_COST, false)
			}
		}

		return true
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return EnergyConsumerItem(stack, MAX_ENERGY)
	}

	override fun getAttributeModifiers(
		slot: EquipmentSlot,
		itemStack: ItemStack
	): Multimap<Attribute, AttributeModifier> {
		if (slot != EquipmentSlot.MAINHAND) {
			return ImmutableMultimap.of()
		}

		val energy = itemStack.matteryEnergy

		if (energy != null && energy.batteryLevel >= ENERGY_PER_SWING) {
			return chargedAttributes
		}

		return dischargedAttributes
	}

	companion object {
		private val MAX_ENERGY = ImpreciseFraction(400_000)
		private val ENERGY_ZAP = ImpreciseFraction(10_000)
		private val ENERGY_PER_SWING = ImpreciseFraction(4_000)
		private val COBWEB_POWER_COST = ImpreciseFraction(1_500)
		private val PLANT_POWER_COST = ImpreciseFraction(500)

		private val DESCRIPTION = TranslatableComponent("item.overdrive_that_matters.energy_sword.desc").withStyle(ChatFormatting.DARK_GRAY)
		private val DESCRIPTION2 = TranslatableComponent("item.overdrive_that_matters.energy_sword.desc2").withStyle(ChatFormatting.DARK_GRAY)
	}
}
